#include<sys/types.h>
#include<sys/socket.h>
#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<stralloc.h>
#include<netdb.h> // gethostbyname
#include<unistd.h>
#include "readlinecrlf.h"

#define MAXREDIRS (20)
#define MAXURLSIZE (5000)

int visited=0;
char visitedURLs[MAXREDIRS][MAXURLSIZE];

// Markiere ein URL als besucht. War der URL schon besucht
// oder ist die maximale Umleitungstiefe MAXREDIRS erreicht,
// ist der Rueckgabewert 0, sonst 1.
int visit(char* url) {
	for(int i = 0; i < visited; i++) 
		if(strcmp(visitedURLs[i], url) == 0) 
			return 0;
	if(visited==MAXREDIRS) 
		return 0;
	strncpy(visitedURLs[visited++], url, MAXURLSIZE-1);
	return 1;
}

// Zerteilt einen URL in Hostname, Portnummer und Dokument
// Dabei werden jeweils maximal maxlen Zeichen geschrieben.
// Rueckgabewert hat stralloc-Semantik:  1, falls erfolgreich,
// sonst 0
int spliturl(char* url, char* hostname, int* port, char* document, int maxlen) {
	// Zeigt Anfang und Ende des Hostnamens an
	char* begin = 0;
	char* end = 0;
	char* current;
	char* last;
	char* portstr = 0;
	char* endofhostname;
	
	for(current = url; (*current)!=0; last=current++) {
		if(*current == ':' && begin!=0) {
			portstr = ++current;
		}
		if(*current == '/') {
			if(begin==0) {
				if(*last == '/') {
					// Doppelslash "//" von "http://"
					begin = ++current;
					continue;
				}
			} else if(end==0) {
				end = current;
				break;	
			}
		}
	}

	if(portstr == 0) {
		*port = 80;
		endofhostname = end;
	} else {
		*port = atoi(portstr);
		if(*port<=0)
			return 0;
		// portstr-1 geht ab dem Doppelpunkt vorm Port los
		endofhostname = portstr-1;
	}
	if(*port > 65535)
		return 0;
	if(begin==0)
		return 0;
	if(end==0)
		return 0;
	if(endofhostname-begin+1 > maxlen) 
		return 0;
	if(strlen(end) > maxlen)
		return 0;

	
	while(begin!=endofhostname) 
		*hostname++ = *begin++;
	*hostname = 0;

	// Leerer Body der Schleife
	while((*document++ = *end++)) ;
	return 1;		
}

int fetch(char* url) {
	char hostname[MAXURLSIZE];
	char document[MAXURLSIZE];
	int port;
	if(!spliturl(url, hostname, &port, document, MAXURLSIZE)) {
		fprintf(stderr, "Der URL %s konnte nicht geparst werden!\n", url);
		return 0;
	}
	printf("URL erfolgreich geparst.\nHostname: %s\nPort: %d\nDokument: %s\n", hostname, port, document);

	if(!visit(url)) {
		fprintf(stderr, "Umleitung im Kreis oder maximale Umleitungstiefe erreicht. Breche ab.\n");
		return 0;
	}	

	struct sockaddr_in addr;
	struct hostent* hp;
	int socketfd;

	if(!(hp = gethostbyname(hostname))) {
		fprintf(stderr, "Unbekannter Host: %s\n", hostname);
		return 0;
	}
	
	// das Adress-Struct ausnullen
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET; // TCP/IPv4-Verbindung
	addr.sin_port = htons(port); // Umwandlung in Network Byte Order
	memcpy(&addr.sin_addr, hp->h_addr, hp->h_length); // Adresse kopieren
	
	if((socketfd=socket(AF_INET,SOCK_STREAM,0)) < 0) {
		perror("socket");
		return 0;
	}
	
	if(connect(socketfd, (struct sockaddr*)&addr, sizeof(addr)) == -1) {
		perror("connect");
		return 0;
	}

	char request[MAXURLSIZE+1000];
	sprintf(request, "GET %s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", document, hostname);
	int requestlen = strlen(request);
	int remaining = requestlen;
	while(remaining > 0) {
		int written = write(socketfd, request+requestlen-remaining, remaining);
		if(written == -1) {
			perror("write");
			return 0;
		}
		remaining -= written;
	}

	stralloc sa = {0};
	stralloc_ready(&sa, BUFSIZ);
	stralloc_copys(&sa, "");
	inbuf buf = {0};
	buf.fd = socketfd;
	inbuf_alloc(&buf, BUFSIZ);

	readlinecrlf(&buf, &sa);
	stralloc_0(&sa);
	int statuscode = atoi(sa.s+9); // HTTP/1.1 ueberspringen
	if(statuscode < 100) {
		fprintf(stderr, "Konnte Statuscode nicht aus Antwort \"%s\" lesen.\n", sa.s);
		return 0;
	}
	printf("Antwort: %s, Statuscode: %d\n", sa.s, statuscode);

	if(statuscode == 301 || statuscode == 302) {
		printf("Das Dokument befindet sich woanders.\n");
		char* loc = "Location: ";
		int loclen = strlen(loc);
		while(sa.len > 1 && strncmp(loc, sa.s, loclen)!=0) {
			readlinecrlf(&buf, &sa);
			stralloc_0(&sa);
		}
		// Entweder wurde der Location-Header gefunden, oder die Header sind zu ende.
		if(sa.len == 1) {
			fprintf(stderr, "Location-Header nicht gefunden. Breche ab.\n");
			return 0;
		}
		printf("Neuer URL: %s\n", sa.s+loclen);
		return fetch(sa.s+loclen);
	}

	if(statuscode!=200) {
		printf("Statuscode ist nicht 200. Gebe Body nicht aus.\n");
		return 0;
	}

	while(sa.len > 0) {
		// So lange Header ueberspringen, bis der Body kommt
		readlinecrlf(&buf, &sa);
	}
	
	int read;
	char answerbuf[BUFSIZ];
	while((read = inbuf_read(&buf, &answerbuf, sizeof(answerbuf))) > 0) {
		int sum=0;
		while(sum < read) {
			int written = write(1, answerbuf, read-sum);
			if(written == -1) {
				perror("write");
				return 0;
			}
			sum += written;
		}
	}
	if(read == -1) {
		perror("read");
		return 0;
	}

	return 1;
}

int main(int argc, char* argv[]) {
	if(argc<2) {
		fprintf(stderr, "Usage: %s url\n", argv[0]);
		exit(255);
	}

	if(!fetch(argv[1])) {
		fprintf(stderr, "Fehler beim Holen eines URL\n");
		return 1;
	}

	return 0;
}
